/*
 * $Id:$
*
 * Copyright (c) 2005 Ivo Hazmuk
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *        This product includes software developed by Emmanuel Dreyfus
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/time.h>

#include <netinet/in.h>
#include <arpa/inet.h>

#include <errno.h>
#include <fcntl.h>
#include <netdb.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <stdarg.h>
#include <termios.h>
#include <unistd.h>

#include <tcpd.h>

#define BUFLEN	1024
#define STRLEN  32
#define VERSION "0.10"
#define RELEASE "20050425"
#define PARAMSTR "a:d:Df:FhH:p:P:s:V"

int		allow_severity = LOG_INFO;
int		deny_severity = LOG_WARNING;
fd_set	rfd, wfd, efd;				/* File descriptors for select */
char	temp[STRLEN] = "NO DATA";	/* Buffer for serial port */
int		debug = 0;					/* Level of debugging information */
pid_t	pid, ppid, chpid;			/* my, parent PID and PID from fork() */
char	pidfile[BUFLEN] = "/var/run/ted.pid";	/* PID filename */
int		daem_b = 1;

struct	request_info request;		/* TCP wrappers request */
int		sock = 0;					/* sockets */
unsigned int	sleeptime = 1;		/* if access denied sleep sleeptime sec */

int		serport;					/* FD for serial device */
int		status, okg, st2;
long	baud;

struct termios tp, tnew;
static char comfname[BUFLEN] = "/dev/cuaa0"; /* default serial port */
int		port_b = 0;					/* inet port override /etc/services */
u_short	port = 3111;				/* default port number; SHOW /etc/services */

/* 
 * Print help screen
 */
void help()
{

	fprintf(stdout, "usage: %s <parameters>\n", getprogname());
	fprintf(stdout, "  parameters: -a<IP> - IP address which is bind\n");
	fprintf(stdout, "              -d# - set debug level, default = 0\n");
	fprintf(stdout, "              -D  - daemonize (default)\n");
	fprintf(stdout, "              -f<name> - set input fname, default %s\n", \
		comfname);
	fprintf(stdout, "              -F  - don't daemonize.\n");
	fprintf(stdout, "              -h  - this help screen\n");
	fprintf(stdout, "              -p# - listening port number.\n");
	fprintf(stdout, "              -P<name> - PID filename, default %s.\n", \
		pidfile);
	fprintf(stdout, "              -s# - if wrapper denied access sleep # second.\n");
	fprintf(stdout, "              -V  - print version information\n");
	if (debug) fprintf(stdout, "Debug level= %d\n", debug);
}
/* end of help */

/*
 * print version information
 */
void version()
{

	fprintf(stdout, "program: %s, ver. %s, Release: %s\n", getprogname(), \
		VERSION, RELEASE);
	if (debug) fprintf(stdout, "%s: at line %d\n", __func__, __LINE__);
}
/* end of version */

/*
 * Set serial line communication parameters
 */
int set_port() {
	tcflush(serport,TCIFLUSH);
	/* set serial port at 9600, 8 bits, 2 stopbits */
	if (tcgetattr(serport, &tp) < 0) {
		syslog(LOG_ERR, "couldn't get term attributes\n");
		return -1;
	}

	/* set controlling flags */
	bzero(&tnew,sizeof(tnew));
	/* CS8 = 8 bits; CLOCAL = ignore control signals */
	baud=B9600;
	tnew.c_cflag = CS8|CLOCAL|CREAD;

	/* settings output flags */
	tnew.c_oflag = 0;
	/* ignore parity */
	tnew.c_iflag = IGNPAR;
	tnew.c_lflag = 0;
	tnew.c_ispeed = B9600;
	tnew.c_ospeed = B9600;
  
	tcflush(serport,TCIFLUSH);
	if (tcsetattr(serport, TCSANOW, &tnew) < 0) {
		syslog(LOG_ERR, "couldn't set term attributes");
		return -1;
	}

	return 0;
}
/* end of set_port */

/*
 * Set some output signals to - voltage -> power off thermometer
 */
int power_off()
{
	int oks;

	status &= ~TIOCM_DSR;
	status &= ~TIOCM_DTR;
	status &= ~TIOCM_RTS;
	oks = ioctl(serport, TIOCMSET, &status);
	
	if (debug >= LOG_DEBUG)
		syslog(LOG_DEBUG, "DSR set off: %x (%d .. %d) \n",status, oks, errno);

	okg = ioctl(serport, TIOCMGET, &st2);
	if (debug >= LOG_DEBUG)
		syslog(LOG_DEBUG, "status: %x (%d .. %d) \n",st2, okg, errno);

	return oks;
}
/* end of power_off */

/*
 * Set some output signals to + voltage -> power on the thermometer
 */
int power_on()
{
	int oks;

	status |= TIOCM_DSR;
	status |= TIOCM_DTR;
	status |= TIOCM_RTS;
	oks = ioctl(serport, TIOCMSET, &status);

	if (debug >= LOG_DEBUG)
		syslog(LOG_DEBUG, "DSR set off: %x (%d .. %d) \n",status, oks, errno);

	okg = ioctl(serport, TIOCMGET, &st2);
	if (debug >= LOG_DEBUG)
		syslog(LOG_DEBUG, "status: %x (%d .. %d) \n",st2, okg, errno);

	return oks;
}
/* end of power_on */

int wrapper(u_int32_t ipaddr)
{
	char client[STRLEN] = "host.domain.cz";
	char client_addr[STRLEN] = "192.168.3.24";
	char client_user[STRLEN] = "";
	char daemon[STRLEN] = "ted";
	int verdict = 1;

	if (debug >= LOG_DEBUG) syslog(LOG_DEBUG, "%s: begin\n", __func__);

	snprintf(client_addr, STRLEN-1, "%d.%d.%d.%d", 
		(ipaddr & 0xff000000) >> 24, (ipaddr & 0x00ff0000) >> 16, \
		(ipaddr & 0x0000ff00) >> 8, ipaddr & 0x000000ff);

	verdict = hosts_ctl(daemon, client_user, client_addr, client_user);
	if (debug >= LOG_DEBUG) syslog(LOG_DEBUG, "%s: at line %d \n", \
		__func__, __LINE__);
	if (debug >= LOG_DEBUG) syslog(LOG_DEBUG, "%s: at line %d \n", \
		__func__, __LINE__);

	if ((debug >= LOG_WARNING) | (verdict == 0)) {
		syslog(LOG_WARNING, "%s: access from %s %s\n", __func__, client_addr, \
		verdict ? "granted" : "denied");
	}
	if (verdict == 0) {
		sleep(sleeptime);
	}

	return verdict;
}
/* end of wrapper */

int printsock(int fd)
{
	socklen_t slen;					/* length of structure nfrom */
	struct sockaddr_in nfrom;		/* */
	struct in_addr ipa;
	u_int32_t ipaddr;

	slen = sizeof(struct sockaddr);
	if (getpeername(fd, (struct sockaddr *)&nfrom, &slen) < 0) {
		syslog(LOG_ERR, "%s: at line %d getpeername(): %d\n", \
		__func__, __LINE__, errno);
	}
	else {
		ipaddr = htonl((u_int32_t)nfrom.sin_addr.s_addr);
		if (debug >= LOG_NOTICE) {
			syslog(LOG_NOTICE, "%s: connect from peer: %d.%d.%d.%d\n", \
			__func__, (ipaddr & 0xff000000) >> 24, (ipaddr & 0x00ff0000) >> 16, \
			(ipaddr & 0x0000ff00) >> 8, ipaddr & 0x000000ff);
		}
		if (wrapper(ipaddr) == 0) {
			return -1;
		}
	}

	return 0;
}
/* end of printsock */

/*
 * Open filedescriptor for socket and send out data from thermometer
 */
int netproc(int fd, int act)
{
	int	rd;
	int	newfd;		/* new socket after accept() */
	struct	sockaddr_in from;
	socklen_t fromlen;

	if (debug >= LOG_DEBUG)
		syslog(LOG_DEBUG, "%s: started processing socket %d\n", __func__, fd);
	switch (act) {
		case 0: 
			newfd = accept(fd, (struct sockaddr *)&from, &fromlen);
			if (newfd >= 0) {
				if (printsock(newfd) != -1) {
					if (debug >= LOG_DEBUG)
						syslog(LOG_DEBUG, "%s: write data to fd: %d\n", \
						__func__, newfd);
				    rd = write(newfd, temp, strlen(temp));
				    rd = write(newfd, "\n", strlen("\n"));
				}
			    close(newfd);
			    shutdown(newfd, SHUT_RDWR);
			}
			else {
			    syslog(LOG_ERR, "%s: at line %d accept(): %d\n", \
				__func__, __LINE__, errno);
			}
			break;
		case 1:
			if (debug >= LOG_NOTICE)
				syslog(LOG_NOTICE, "%s writing\n", __func__);
			break;
		case 2:
			if (debug >= LOG_NOTICE)
				syslog(LOG_NOTICE, "%s exceptions\n", __func__);
			break;
	}
	return 0;
}
/* end of netproc */

void serproc(int fd, int act)
{
	ssize_t din;	/* how many chars I've read */
	int i;

	if (debug >= LOG_DEBUG)
		syslog(LOG_DEBUG, "%s: started processing socket %d\n", __func__, fd);
	switch (act) {
		case 0:
			usleep(100000);	/* wait 0.1s for all data on input */
			din = read(serport, &temp, STRLEN-1);
			for (i=0; i <= din; i++) {
			    if (temp[i] < ' ') temp[i] = 0x0;
			}
			break;
		case 1:
			if (debug >= LOG_NOTICE)
				syslog(LOG_NOTICE, "%s: writing\n", __func__);
			break;
		case 2:
			if (debug >= LOG_NOTICE)
				syslog(LOG_NOTICE, "%s: exceptions\n", __func__);
			break;
	}
	if (debug >= LOG_DEBUG)
		syslog(LOG_DEBUG, "%s: measured temp: %s\n", __func__, temp);
}
/* end of serproc */

/*
 * Read port number from /etc/services
 */
u_short readport()
{
	char const sername[] = "tm";
	char const serproto[] = "tcp";
	struct servent *service;

	if (debug >= LOG_DEBUG)
		syslog(LOG_DEBUG, "%s: started processing\n", __func__);
	if (!port_b) {
		setservent(0);
		if (debug >= LOG_DEBUG)
			syslog(LOG_DEBUG, "setservent at line %d \n", __LINE__);
		service = getservbyname(sername, serproto);
		if (debug >= LOG_DEBUG)
			syslog(LOG_DEBUG, "getservbyname at line %d \n", __LINE__);

		if (service != NULL) {
		    if (debug >= LOG_DEBUG)
				syslog(LOG_DEBUG, "%s: we will listen on port %d\n", \
				__func__, ntohs(service->s_port));
		    port = ntohs((u_short)service->s_port);
		}
		else {
		    syslog(LOG_ERR, "Cannot found service \"%s\" in /etc/services\n", sername);
		}
	}

	return port;
}
/* end of readport */

/*
 * Create and fill PID file.
 */
int write_pid(pid_t pid)
{
	int fd;
	char buf[BUFLEN];

	fd = open(pidfile, O_CREAT | O_TRUNC | O_WRONLY);
	if (fd < 0) {
		syslog(LOG_ERR, "%s: cannot write PID file %s.\n", __func__, pidfile);
	}
	snprintf(buf, BUFLEN-1, "%d\n", pid);
	write(fd, buf, strlen(buf));
	close(fd);
}
/* end of write_pid */

/*
 * Close sockets, remove PID file and terminate daemon.
 */
void terminates(int sig)
{
	int	ok;
	
	syslog(LOG_NOTICE, "%s: daemon exited with signal %d\n", __func__, sig);

	if (shutdown(sock, SHUT_RDWR) == -1) syslog(LOG_ERR, \
		"%s: problem to shutdown socket %d\n", __func__, sock);
	if (close(sock) == -1) syslog(LOG_ERR, \
		"%s: problem to close socket %d\n", __func__, sock);

	ok = power_off();
	if (debug >= LOG_NOTICE) syslog(LOG_NOTICE, \
		"%s: thermometer powered off\n", __func__);
	if (unlink(pidfile) == -1) syslog(LOG_ERR, \
		"%s: unlink() at line %d failed with error %d\n", __func__, \
		__LINE__, errno);

	exit(EXIT_SUCCESS);
}
/* end of terminates */

/*
 * Internet server sending to client values measured by the thermometer
 * from Papouch s.r.o.
 * See http://www.papouch.com/
 */
int main(int argc, char **argv)
{
int		nfd;			/* number of fd ready for processing */
int		ok;
u_int32_t ipaddr;
char	ipa[STRLEN];	/* IP address string from command line */
struct	in_addr pin;
struct	sockaddr_in addr;
struct	timeval tv;
int		help_b = 0;		/* help() is invoked */
int		ver_b = 0;		/* version() is invoked */
int		adr_b = 0;		/* IP address set from command line */
char	hostname[BUFLEN];

	setprogname(argv[0]);
	openlog(getprogname(),LOG_PID,LOG_LOCAL7);
	ipaddr = INADDR_ANY;
	syslog(LOG_NOTICE, "started.\n");
	/*
	 * Parse command line
	 */
	while ((ok = getopt(argc, argv, PARAMSTR)) != -1) {
		switch (ok) {
		    case 'a': /* set IP address */
				strncpy(ipa, optarg, STRLEN);
				ok = inet_aton(ipa, &pin);
				if (!ok) {
					syslog(LOG_ERR, "string %s is bad.\n", ipa);
				}
				else {
					adr_b = 1;
					bcopy(&pin, &ipaddr, 4);
				}
				break;
		    case 'd': /* set debug level */
				debug = atoi(optarg);
				if (debug >= LOG_NOTICE) syslog(LOG_NOTICE, \
					"%s: debug level %d\n", __func__, debug);
				break;
			case 'D': /* daemonize obsoleted */
				break;
			case 'f': /* set serial port name */
				strncpy(comfname, optarg, BUFLEN);
				break;
			case 'F': /* don't daemonize */
				daem_b = 0;
				break;
		    case 'h': /* help */
		    case '?':
				help_b = 1;
				break;
		    case 'H': /* I don't know, I've forgotten */
				break;
		    case 'p': /* set inet port number */
				port_b = 1;
				port = atoi(optarg);
				if (port == 0)
					syslog(LOG_ERR, \
					"set port to 0 invoke listening on undefined port!\n");
				if (debug >= LOG_NOTICE)
					syslog (LOG_NOTICE, \
					"port %d set for listening.\n", port);
				break;
			case 'P': /* set PID filename */
				strncpy(pidfile, optarg, BUFLEN);
				break;
			case 's': /* set sleeptime */
				sleeptime = atoi(optarg);
				break;
		    case 'V': /* print version information */
				ver_b = 1;
				break;
		}
	}

	if (ver_b) version();
	if (help_b) help();
	if (help_b || ver_b) exit(EXIT_SUCCESS);

	syslog(LOG_NOTICE, "read data from %s, listen on port %d.\n", \
		comfname, readport());
	/*
	 * Serial line settings
	 */
	if ((serport = open(comfname, O_RDONLY | O_NOCTTY)) < 0) {
		syslog(LOG_ERR, "Exiting. Cannot open serial port\n");
		exit(EXIT_FAILURE);
	}
	if (debug >= LOG_DEBUG) syslog(LOG_DEBUG, "serial socket: %d\n", serport);

	ok = set_port();
	if (ok < 0) syslog(LOG_ERR, "set_port return %d\n");
	ok = power_off();
	sleep(1);		/* wait 1 second for shutdown TM */
	ok = power_on();

	/*
	 * Internet server settings
	 */
	if (debug >= LOG_NOTICE) {
		syslog(LOG_NOTICE, "IP: %d.%d.%d.%d\n", \
			(ipaddr & 0xff000000) >> 24, (ipaddr & 0x00ff0000) >> 16, \
			(ipaddr & 0x0000ff00) >> 8, ipaddr & 0x000000ff);
	}

	if ((sock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0) {
		syslog(LOG_ERR, "Exiting. Cannot open Internet socket\n");
		exit(EXIT_FAILURE);
	}
	else {
		if (debug >= LOG_NOTICE)
			syslog(LOG_NOTICE, "network socket: %d\n", sock);
		bzero(&addr, sizeof(addr));
		if (debug >= LOG_DEBUG)
			syslog(LOG_DEBUG, "bzero at line %d \n", __LINE__);
		addr.sin_family = AF_INET;
		if (debug >= LOG_DEBUG) 
			syslog(LOG_DEBUG, "addr.sin_family at line %d \n", __LINE__);
		addr.sin_port = htons(readport());
		if (debug >= LOG_DEBUG)
			syslog(LOG_DEBUG, "readport at line %d \n", __LINE__);
		bcopy(&ipaddr, &addr.sin_addr, 4);
		if (debug >= LOG_DEBUG)
			syslog(LOG_DEBUG, "bcopy at line %d \n", __LINE__);
		if (debug >= LOG_NOTICE)
			syslog(LOG_NOTICE, "Addr: %x\n", addr.sin_addr);
	}
	if (bind(sock, (struct sockaddr *)&addr, sizeof addr) < 0) {
		syslog(LOG_ERR, "Exiting. Bind error %d\n", errno);
		exit(EXIT_FAILURE);
	}
	listen(sock, 2);

	/*
	 * Set timeout for checkig file descriptors by select().
	 * I know, that serial port send data each 10 seconds.
	 */
	tv.tv_sec = 12;
	tv.tv_usec = 0;

	/*
	 * Create child process which state daemon.
	 * Parent process will exit immediately.
	 */
	if (daem_b) {
		chpid = fork();
		if (chpid == -1) {
			syslog(LOG_ERR, "cannot fork\n");
		}
		if (chpid == 0) {
			if (daemon(1,0) == -1) {
				syslog(LOG_ERR, "error %d in daemon() at line %d.\n", \
					errno, __LINE__);
				syslog(LOG_ERR, "exiting!!!\n");
				exit(EXIT_FAILURE);
			}
			syslog(LOG_NOTICE, "child continue.\n");
		}
		else {
			return 0;
		}
	}

	write_pid(getpid());
	signal(SIGTERM, &terminates);
	signal(SIGHUP, &terminates);
	signal(SIGKILL, &terminates);

	/*
	 * TCP wrappers
	 */
	gethostname(hostname, BUFLEN);
	request_init(&request, RQ_DAEMON, getprogname(), RQ_SERVER_NAME, hostname, \
	0);

	/*
	 * Neverending loop which make interaction with whole world
	 */
	while (1) {
		/* set file desciptors for select() */
		FD_ZERO(&rfd);
		FD_SET(sock, &rfd);
		FD_ZERO(&wfd);
		FD_SET(sock, &wfd);
		FD_ZERO(&efd);
		FD_SET(sock, &efd);
		FD_SET(serport, &rfd);
		FD_SET(serport, &efd);

		if (debug >= LOG_DEBUG) syslog(LOG_DEBUG, "will be select()\n");
		nfd = select((1 + sock), &rfd, &wfd, &efd, &tv);
		switch (nfd) {
			case -1:
			    syslog(LOG_ERR, "select() error %d.\n", errno);
			    break;
			case 0:
			    if (debug >= LOG_DEBUG)
					syslog(LOG_DEBUG, "select timeout.\n");
			    break;
			default:
			{
			    if (debug >= LOG_DEBUG)
					syslog(LOG_DEBUG, "%d descriptors ready.\n", nfd);
			    if (FD_ISSET(sock, &rfd)) {
			    	netproc(sock,0);
			    }
			    if (FD_ISSET(sock, &wfd)) {
					netproc(sock,1);
			    }
			    if (FD_ISSET(sock, &efd)) {
					netproc(sock,2);
			    }
			    if (FD_ISSET(serport, &rfd)) {
					serproc(sock,0);
			    }
			    if (FD_ISSET(serport, &wfd)) {
					serproc(sock,1);
			    }
			    if (FD_ISSET(serport, &efd)) {
					serproc(sock,2);
			    }
			}
		}
	};
	closelog();


	return 0;
}

